#!/bin/sh
#
# Verify external storage medium and mount it
#
# Note: It is required that the external storage is not automatically
# mounted by the mountall boot step (fstab: noauto).

minpartsize=$((128 * 1024))
REBOOT_LOG=/var/log/sdcard/reboots.log
FSCK_LOG=/var/log/sdcard/fsck.log

# read by dss11-websetup/www/cgi-bin/libsdcard.sh
FSCK_STATUS=/var/run/sdcard-fsck.status

isnum() {
    # http://stackoverflow.com/questions/2210349/bash-test-whether-string-is-valid-as-an-integer
    [ -n "$1" -a $1 -eq $1 ] 2> /dev/null || {
        # if $1 is neither empty nor a number it would return '2'
        return 1
    }
}

emit_log() {
    echo "SD-CARD: $*"
}

request_reboot()
{
    # ensure file exists and we have permissions
    # if we err' here we can create a nasty reboot loop
    touch $REBOOT_LOG
    err=$?
    if [ $err -ne 0 ]; then
        emit_log "Can't access $REBOOT_LOG, ignore reboot request"
        return 250
    fi

    ct=`wc -l $REBOOT_LOG | cut -f1 -d\ `
    echo "dbg> ct: \"$ct\""

    if isnum $ct; then
        emit_log "Failed to count lines of $REBOOT_LOG", ingore reboot request
        exit 249
    fi

    echo "number of reboots: $ct"
    if [ $ct -gt 5 ]; then
        emit_log "Too many reboots requested, continue without card"
        return 255
    fi

    emit_log "schedule reboot"
    date +"%Y-%m-%dT%H:%M:%S" >> $REBOOT_LOG
    shutdown -r now
}


fetch_blkid_desc()
{
    # will define LABEL / UUID / TYPE variable
    eval `blkid /dev/mmcblk0p1 | cut -f2 -d:`
}

format_sd()
{
    error=0 # set -e, likes not errors
    emit_log "creating ext2 filesystem..."
    echo 1 > /sys/class/leds/blue/brightness
    echo 0 > /sys/class/leds/green/brightness
    echo 1 > /sys/class/leds/red/brightness

    exec 7>&1
    exec 8>&2
    exec >> $FSCK_LOG 2>&1
    mkfs.ext2 -v -L DSSDATA $sdoptions /dev/mmcblk0p1 || error=$?
    exec 1>&7 7>&-
    exec 2>&8 8>&-
    echo $error > $FSCK_STATUS
    return $error
}

fsck_sd()
{
    mode=$1
    error=0 # set -e, likes not errors

    exec 7>&1
    exec 8>&2
    exec >> $FSCK_LOG 2>&1
    if [ "x$mode" = "xautorepair" ]; then
        emit_log "[fsck autorepair]"
        fsck.ext4 -v -y -f /dev/mmcblk0p1 || error=$?
        echo $error > $FSCK_STATUS
    else
        fsck.ext4 -v -p /dev/mmcblk0p1 || error=$?
        echo $error > $FSCK_STATUS
    fi
    exec 1>&7 7>&-
    exec 2>&8 8>&-

    emit_log "fsck status: $error"
    return $error
}

has_journal()
{
    dumpe2fs -h /dev/mmcblk0p1 2> /dev/null | \
        grep -qe "^Filesystem\ features:[ \t]*has_journal"
}

disable_journal()
{
    tune2fs -O ^has_journal /dev/mmcblk0p1
}

mount_sd()
{
    FORMAT=0
    sdoptions=''

    if [ ! -b /dev/mmcblk0 ]; then
        emit_log "card not found, aborting... "
        request_reboot
        return 254
    fi

    # fdisk will trigger update of /sys/block/mmcblk0
    if [ ! -b /dev/mmcblk0p1 ]; then
        emit_log "invalid partition table: re-create..."
        # new partition, primary, partition 1, default, default, w
        echo "n
        p
        1


        w
        "| fdisk /dev/mmcblk0
        sleep 5 # without sleep, mke2fs will not find p1
        FORMAT=1
    fi

    if [ ! -d /sys/block/mmcblk0/mmcblk0p1 ]; then
        emit_log "meta information not available, aborting..."
        request_reboot
        return 253
    fi

    fetch_blkid_desc
    if ((test -n "$TYPE") &&
        (test "$TYPE" != "ext4") &&
        (test "$TYPE" != "ext3") &&
        (test "$TYPE" != "ext2")) ; then
        emit_log "wrong filesystem: \'$TYPE\' will reformat"
        FORMAT=1
    fi

    sdsize=$(cat /sys/block/mmcblk0/mmcblk0p1/size)
    if [ $sdsize -le $minpartsize ]; then
        # TODO maybe card is partitioned wrong, check size of card
        emit_log "partition too small $sdsize kB, ignore device.... "
        return 255
    fi

    if [ -f /etc/.sdformat ]; then
        FORMAT=1
        sdoptions=`head -n 1 /etc/.sdformat` # bogus, see cgi-bin/format.sh
        emit_log "user requested reformat with options: \"$sdoptions\""
        # mkfs.ext4 options, eg. "-cc" for intensive bad block tests
    fi

    if [ $FORMAT -eq 0 ]; then
        emit_log "start filesystem check..."
        error=0 # set -e likes no errors
        fsck_sd || error=$?

        if [ $(($error & 4)) == 4 ] ; then
            emit_log "automatic repair filesystem"
            fsck_sd autorepair
            error=$?
        fi

        if [ $(($error & 2)) == 2 ] ; then
            emit_log "fsck requested reboot"
            request_reboot
            return 126
        fi

        if [ $(($error & ~3)) != 0 ] ; then
            emit_log "Fatal errors on SD-CARD detected [fsck:$error]"
            FORMAT=1
        fi
    fi

    # aggregate .sdformat/re-partition/blkid format requests
    if [ $FORMAT -eq 1 ]; then
        format_sd
        if [ $? -ne 0 ]; then
            emit_log "WARNING failed to format SD-CARD"
            request_reboot
            return 125
        fi

        # paranoia
        fsck_sd
        error=$?
        if [ $error -ne 0 ]; then
            emit_log "WARNING new filesystem has errors"
            request_reboot
            return 124
        fi
        rm -f /etc/.sdformat
    fi

    if has_journal; then
        # journal increases number of writes, disable
        emit_log "Disable journal on sdcard"
        disable_journal
    fi

    mount /media/card

    # fresh start
    echo -n > $REBOOT_LOG
}

prepare_overlay()
{
    #
    #  /etc/fstab is maintained by base-files package
    #  we need to create the branch directory for those
    #  mounts to work
    #
    #  mkdir -p /media/card/sys/log
    #  mkdir -p /media/card/sys/metering
    #
    emit_log "creating filesystem overlays"
    while read DIR; do
        # echo "  $DIR"
        mkdir -p $DIR
    done <<EOF
    $(grep ^aufs /etc/fstab | sed -n 's/.*,br:\([^,=]*\).*/\1/p')
EOF
}

streamcopy ()
{
    source=$1
    dest=$2
    options=$3
    [ -d $dest ] || mkdir -p $dest
    (cd $source; tar -cf - --exclude=".*" $options * | tar -C $dest -xf -)
}

move2internal ()
{
    echo -n "SD-CARD: merging data to internal storage: "

    #
    # transitional step:
    # previously, most user data was stored on sd-card
    # move those data back into more reliable internal
    # flash. executed only once per installation
    #

    if [ -e /media/card/ssl ]; then
        echo -n "Certificates "
        streamcopy /media/card/ssl/ /etc/ssl/certs/
        mv /media/card/ssl /media/card/ssl.moved
    fi

    if [ -e /media/card/sys/dss/data/metering ]; then
        echo -n "Metering "
        streamcopy /media/card/sys/dss/data/metering/ /media/card/sys/metering/
        rm -rf /media/card/sys/dss/data/metering
    fi

    if [ -e /media/card/sys/dss/data ]; then
        echo -n "dSS-Runtime "
        streamcopy /media/card/sys/dss/data/ /usr/share/dss/data/
        mv /media/card/sys/dss/data /media/card/sys/dss/data.moved
    fi

    if [ -e /media/card/sys/dss/add-ons ]; then
        echo -n "dSS-Addons "
        streamcopy /media/card/sys/dss/add-ons/ /usr/share/dss/add-ons/
        mv /media/card/sys/dss/add-ons /media/card/sys/dss/add-ons.moved
    fi

    if [ -e /media/card/sys/add-ons ]; then
        echo -n "Addons "
        streamcopy /media/card/sys/add-ons/ /www/pages/add-ons/
        mv /media/card/sys/add-ons /media/card/sys/add-ons.moved
    fi

    if [ -e /media/card/sys/dsa ]; then
        echo -n "dSA-Database (ignored)"
        #streamcopy /media/card/sys/dsa/ /var/lib/dsa
        #mv /media/card/sys/dsa /media/card/sys/dsa.moved
    fi

    echo ""
    sync
}

if [ -n "$UNITTEST" ]; then
    # exit will abort the unit test
    return
fi

#
# MAIN
#

#
# The opkg install hook runs while the sd-card is mounted. That means
# /var/log/sdcard folder is created on the sd-card. But the sd-card is
# not yet mounted. Ensure that folder exists or hell breaks loose
#
mkdir -p /var/log/sdcard

#
# in fstab it's noauto, so it shouldn't be mounted at startup
# but this script will try fsck, so better safe than sorry
#
umount /dev/mmcblk0p1 > /dev/null 2>&1

#
# will partition, format, mount the sd card. incl. retries
# and reboots. skip if card already mounted
#
if ! mountpoint -q /media/card; then
    mount_sd
fi

if mountpoint -q /media/card; then
    fetch_blkid_desc
    emit_log "volume: $LABEL UUID: $UUID "
    move2internal

    # check and prepare overlay mounts
    prepare_overlay
    mount -a -t aufs || error=$?
fi
