#!/bin/sh

if [ ! -f run_tests.sh ]; then
    echo "Invalid directory. Please run as ./run_tests.sh from the test/ directory."
    exit 1
fi

usage() {
    echo "Usage:"
    echo "  $0 [options] [test_directory]..."
    echo "Options:"
    echo "  -h"
    echo "     Show this help message."
    echo "  -v"
    echo "     Verbose mode"
    echo "  -k"
    echo "     do not stop on failure"
    echo "  -l"
    echo "     list available test cases"
    echo "  -u"
    echo "     Update the test expected output with the new output in case of failure"
}

VERBOSITY_FLAGS=
DRY_MODE_FLAG=

while getopts "hdvkulm:" option; do
  case $option in
    h)
      usage
      exit 0
      ;;
    d)
      DRY_MODE_FLAG="-d"
      ;;
    v)
      VERBOSITY_FLAGS=" $VERBOSITY_FLAGS -v"
      ;;
    k)
      CONTINUE_ON_FAILURE=1
      ;;
    u)
      MOVE_TEST_OUTPUT=1
      ;;
    l)
      find . -type d -name test_\* | cut -c 3-
      exit 0
      ;;
    \?)
      usage
      exit 1
      ;;
  esac
done

shift $(($OPTIND - 1))

if [ "$#" -eq 0 ]; then
    TESTS=$(find . -type d -name test_\* | sort)
else
    TESTS="$@"
fi

TMP_DIR="$(mktemp -d)"

echo "Building source..."
( cd .. &&
  autoreconf -i &&
  ./configure \
    --prefix=/\
    --exec_prefix=/usr\
    --datarootdir=/usr/share ) > "$TMP_DIR/build_output" 2>&1
if [ $? -ne 0 ]; then
    cat "$TMP_DIR/build_output"
    exit 1
fi

install_sysroot()
{
    cp -a --dereference "$1"/* "$2"
}

remove_sysroot()
{
    SRC=$1
    find "$1" -empty -type d | sed -e "s%$SRC/%%g" |
    while read dir_name; do rm -R "$TMP_DIR/sand box/$dir_name"; done
    find "$1" -type f | sed -e "s%$SRC/%%g" |
    while read file_name; do rm "$TMP_DIR/sand box/$file_name"; done
}

populate_sysroot()
{
    local TEST="$1"

    mkdir -p "$TMP_DIR/sand box"
    mkdir -p "$TMP_DIR/sand box/tmp"
    mkdir -p "$TMP_DIR/sand box/mnt"

    cp $TEST/cmdline $TMP_DIR

    ( cd .. && make install DESTDIR="$TMP_DIR/sand box/") > \
        "$TMP_DIR/install_log" 2>&1
    if [ $? -ne 0 ]; then
        echo "make install failed"
        cat "$TMP_DIR/install_log"
    fi

    install_sysroot skeleton "$TMP_DIR/sand box/"
    if [ -d "$TEST"/missing_files ]; then
        remove_sysroot "$TEST"/missing_files
    fi
    if [ -d "$TEST"/files ]; then
        install_sysroot "$TEST"/files "$TMP_DIR/sand box/"
    fi
}

TST_OUTPUT="$TMP_DIR/output"
SDO_ROOT_DIR="sand box" # whitespace sucks :-)
export SDO_ROOT_DIR

echo "Test dir: $TMP_DIR"
PATH="$TMP_DIR/sand box/sbin:$TMP_DIR/sand box/bin:$TMP_DIR/sand box/usr/bin:$TMP_DIR/sand box/usr/sbin:$PATH"
export PATH
for TEST in $TESTS; do
    DESC="${TEST#./}"
    printf "%-50s" "$DESC"

    populate_sysroot "$TEST"


    ( cd $TMP_DIR;
      busybox sh cmdline $VERBOSITY_FLAGS $DRY_MODE_FLAG
    ) > "$TST_OUTPUT" 2>&1 < /dev/null

    EXPECTED="${TEST}"/expected

    DIFF_OUTPUT="$(diff --unidirectional-new-file "$EXPECTED" "$TST_OUTPUT" 2>&1)"
    DIFF_STATUS="$?"
    if [ "$DIFF_STATUS" -ne 0 ] && [ -n "$MOVE_TEST_OUTPUT" ]; then
        mv "$TST_OUTPUT" "$EXPECTED"
    fi
    /bin/rm -Rf "$TMP_DIR/sand box"
    if [ "$DIFF_STATUS" -eq 0 ]; then
        echo " passed."
    else
        echo " failed!"
        if [ -n "$MOVE_TEST_OUTPUT" ]; then
            echo " Updated expected file with the new output."
        fi
        echo "Output of diff $EXPECTED $TST_OUTPUT"
        echo "$DIFF_OUTPUT"
        if [ -z "$CONTINUE_ON_FAILURE" ]; then
            ( cd .. && make distclean )  >> "$TMP_DIR/build_output" 2>&1
            exit 1
        fi
    fi
    # cleanup if next text uses a different machine
    ( cd .. && make clean  >> "$TMP_DIR/build_output" 2>&1)
done

( cd .. && make distclean ) >> "$TMP_DIR/build_output" 2>&1
rm -Rf "$TMP_DIR"
