#!/bin/sh

set -e

. ./cleanup-handler.sh

# debugging:
# remove 'check' in front of failing command
# 'set -x' changes output and 'check' might fail
# -> disable output checking: expect empty string ""

UNITTEST=1
SDO_ROOT_DIR=/tmp/rootfs
. ./sdo-disable-logs.sh # runs with dash

ROOTFS="${SDO_ROOT_DIR}"

# loopback device 2x free space
CRITICAL_USAGE_PERCENT=70 # critical if usage higher than
CRITICAL_FREE_SPACE=5     # critical if less space available

target()
{
    grep -q ARM926EJ /proc/cpuinfo
}

if ! target; then
    echo "Test can only be run on target"
    exit 1
fi

# /tmp is a ramdisk, redirect to disk if RAM is scarce
mkdir -p /alt_tmp
export TMPDIR=/alt_tmp # non-ramdisk

install_loopback()
{
    if [ $(opkg list-installed 'kernel-module-loop' | wc -l) -eq 0 ]; then
        echo "need kernel-module-loop to mount virtual sdcard, install? yes/no"
        while read a; do
            case $a in
                "yes")
                    opkg install kernel-module-loop
                    break;;
                "no")
                    echo "... quitting"
                    exit;;
            esac
            echo "need kernel-module-loop to mount virtual sdcard, install? yes/no"
        done
    fi
    # not automatically done by mount command
    modprobe loop || die "failed to load kernel-module-loop"
}
install_loopback


# for debugging:
debug_wait()
{
    while read a; do
        [ -n "$a" -a "$a" = 'quit' ] && break
    done
}

test_passed()
{
    echo "PASSED"
}

check()
{
    fct=$1
    exp_ret=$2
    exp_output=$3
    ret=0
    # eval runs in a separate shell, it's environment
    # is not visible from this shell
    output=`eval $fct 2>&1` || ret=$?
    [ $ret -eq $exp_ret ] || die "$fct: return code mismatch [$exp_ret != $ret]"

    if [ -z "$exp_output" ]; then
        return
    fi

    [ "$output" = "$exp_output" ] || \
        die "$fct: output mismatch [\'$exp_output\' != \'$output\']"
}

echo "######################################"
echo "TEST: cleanup-handler"
echo "######################################"
touch /tmp/foo /tmp/bar
cleanup_push "rm /tmp/foo"
cleanup_push "rm /tmp/bar"
cleanup
check 'test -f /tmp/foo' 1 ""
check 'test -f /tmp/bar' 1 ""
test_passed

loopback_disk()
{
    img=$(mktemp)
    mntpnt=$1
    size=$(($CRITICAL_FREE_SPACE * 2 * 1024))  # MB

    # accidental left-overs?
    [ -d "$mntpnt" ] && die "$mntpnt already exists"

    echo "-- create virtual sd-card: $size MB"
    dd if=/dev/zero of=$img bs=1024 count=$size > /dev/null 2>&1
    cleanup_push "rm $img"

    /sbin/mkfs.ext2 -F $img > /dev/null 2>&1
    mkdir $mntpnt
    cleanup_push "rmdir $mntpnt"

    # busybox mount can't loop mount directly
    losetup /dev/loop0 $img
    cleanup_push "losetup -d /dev/loop0"
    mount /dev/loop0 $mntpnt
    cleanup_push "umount $mntpnt"
}

echo "######################################"
echo "TEST: mount read-only overlay and verify it's not writable"
echo "######################################"
loopback_disk $ROOTFS

check "mount_ro_aufs $ROOTFS/foo" 2 "$ROOTFS/foo does not exist"
check "mountpoint -q $ROOTFS/foo" 1 ""
mkdir -p $ROOTFS/foo
check "mount_ro_aufs $ROOTFS/foo" 0 ""
check "mountpoint -q $ROOTFS/foo" 0 ""
check "touch $ROOTFS/foo/bar" 1 ""
test_passed

# cleanup manually
umount $ROOTFS/foo
rm -r $ROOTFS/* || true

echo "######################################"
echo "TEST: exceed usage level, disable log folders to reduce usage"
echo "######################################"
# inherit loopback from previous test

mockup_logs()
{
    logdir=$1
    size=${2:-512} # kBytes
    name=${3:-log}
    mkdir -p $logdir
    dd if=/dev/zero of=$logdir/$name bs=1024 count=$size > /dev/null 2>&1 || true
}

mockup_service()
{
    local service="$1"

    mkdir -p "$ROOTFS"/etc/init.d
    #echo "echo $1 \"\$@\"" > "$ROOTFS"/etc/init.d/"$service"
    echo "#!/bin/true" > "$ROOTFS"/etc/init.d/"$service"
    chmod +x "$ROOTFS"/etc/init.d/"$service"
}

# fill fs till 100%
MANAGED_DAEMONS=""
for i in $(seq -w 20 1 60); do
    MANAGED_DAEMONS="$MANAGED_DAEMONS $i"
    local LOGDIR="$(lookup_logdir $i)"

    mockup_logs "$LOGDIR" 512 log
    mockup_service "$i"

    if [ $(fs_pcent_used) -eq 100 ]; then
        break
    fi
done

check enough_free_space 1 ""
echo "used : $(fs_pcent_used)% | free : $(fs_free_mb) MB"

cleanup_push "rm $STATUS" # created by select_victim
umount_aufs() {
    for D in "$(cat $STATUS)"; do
        local LOGDIR="$(lookup_logdir $D)"
        umount "$LOGDIR" 2> /dev/null || true
    done
}
cleanup_push "umount_aufs"

# stop daemons to make sure no open fd, that would prevent
# the fs from freeing the blocks
select_victim
test_passed

echo "######################################"
echo "TEST: check read-only folders, not writable by root"
echo "######################################"
# inherit loopback from previous test

# verify read-only directory are not writable
old_free=$(fs_free_mb)
old_pcent=$(fs_pcent_used)
echo "before: $old_pcent% | free : $old_free MB"

# try to write to now read-only folders
for D in `cat $STATUS`; do
    local LOGDIR="$(lookup_logdir $D)"
    mockup_logs "$LOGDIR" 1
done

echo "after : $(fs_pcent_used)% | free : $(fs_free_mb) MB"
check fs_free_mb 0 $old_free  # usage didn't change
check fs_pcent_used 0 $old_pcent
test_passed

echo "######################################"
echo "TEST: grow remaining writable logs, then disable more logs to reduce usage"
echo "######################################"
# inherit loopback from previous test

# enlarge still writable logs
for D in $MANAGED_DAEMONS; do
    LOGDIR="$(lookup_logdir $D)"
    if mountpoint -q $LOGDIR; then
        # already disabled
        continue
    fi
    echo "$LOGDIR not yet disabled, grow log files"
    for i in $(seq -w 0 5); do
        mockup_logs "$LOGDIR" 512 addon$i
    done
done

echo "used : $(fs_pcent_used)% | free : $(fs_free_mb) MB"
check enough_free_space 1 ""

# re-run algorithm
select_victim

echo "used : `fs_pcent_used`% | free : `fs_free_mb` MB"
check enough_free_space 0 ""
test_passed

echo "######################################"
echo "TEST: invalid MANAGED_DAEMONS config "
echo "######################################"

# enlarge still writable logs
for D in $MANAGED_DAEMONS; do
    LOGDIR=`lookup_logdir $D`
    if mountpoint -q $LOGDIR; then
        # already disabled
        continue
    fi
    echo $LOGDIR not yet disabled, grow log files
    for i in `seq -w 0 5`; do
        mockup_logs $LOGDIR 512 addon$i
    done
done

MANAGED_DAEMONS_ORIG=${MANAGED_DAEMONS}
MANAGED_DAEMONS="brumbal ${MANAGED_DAEMONS}"

select_victim
test_passed

echo
echo
echo  "     ============="
echo  "     |  SUCCESS  |"
echo  "     ============="
echo
